#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include "mctd.h"
#include "mctd-log.h"
#include "mctd-serial.h"

#define WAIT_DEFAULT 5

static char _recv_buffer[MCTD_AT_BUF_SIZE];
static int _recv_length = 0;

static int copy_valid_command(char *buf, char *end, bool iscrlf)
{
    int valid_length;

    valid_length = end - _recv_buffer;
    valid_length += iscrlf?2:1;

    memcpy(buf, _recv_buffer, valid_length);

    if (valid_length == _recv_length) {
        memset(_recv_buffer, 0, MCTD_AT_BUF_SIZE);
        _recv_length = 0;
    } else {
        memmove(_recv_buffer, &(_recv_buffer[valid_length]),
                                            MCTD_AT_BUF_SIZE - valid_length);
        _recv_length -= valid_length;
        memset(&(_recv_buffer[MCTD_AT_BUF_SIZE - valid_length]),
                                                            0, valid_length);
    }
    return valid_length;
}

static int copy_at_command(char *buf)
{
    char *p;

    p = strstr(_recv_buffer, "\r\n");
    if (p)
        return copy_valid_command(buf, p, true);

    p = strchr(_recv_buffer, '\r');
    if (p)
        return copy_valid_command(buf, p, false);

    p = strchr(_recv_buffer, '\n');
    if (p)
        return copy_valid_command(buf, p, false);

    return -1;
}

static bool has_received_command()
{
    return (_recv_length > 0);
}

static int read_line_with_timeout(int fd, char *buf,
                                            int timeout_sec, int timeout_usec)
{
    int current_length = 0;

    if ((_recv_length > 0) &&
        (strchr(_recv_buffer, '\r') || strchr(_recv_buffer, '\n'))) {
        return copy_at_command(buf);
    }

    while (mctd_wait_data_with_usec(fd, timeout_sec, timeout_usec)
                                                        == MCTD_RESULT_OK) {
        current_length = read(fd, &(_recv_buffer[_recv_length]),
                                        MCTD_AT_BUF_SIZE - 1 - _recv_length);
        if (current_length < 0) {
            perror("read");
            log_info("read failed.");
            return -1;
        }

        if (current_length == 0) {
            log_info("read:no data");
            return -1;
        }

        _recv_length += current_length;

        if (strchr(_recv_buffer, '\r') || strchr(_recv_buffer, '\n'))
            return copy_at_command(buf);
    }

    return -1;
}

static int read_line(int fd, char *buf)
{
    return read_line_with_timeout(fd, buf, 1, 0);
}

int mctd_atcmd_send(int fd, const char *command, const char *key,
                                            char **response, int *wait_second)
{
    int read_size;
    int result = MCTD_RESULT_ERR_TIMEOUT;
    int retry_remain;
    char buf[MCTD_AT_BUF_SIZE];

    /* dummy read */
    memset(buf, 0, MCTD_AT_BUF_SIZE);
    do {
        read_size = read_line_with_timeout(fd, buf, 0, 1);
        if (read_size > 0)
            log_at_recv(buf);
    } while(read_size > 0);

    if (wait_second) {
        if (*wait_second <= 0) {
            log_info("timed out.");
            sleep(1);
            return MCTD_RESULT_ERR_TIMEOUT;
        }
        retry_remain = *wait_second;
    } else
        retry_remain = WAIT_DEFAULT;

    /* send command */
    sprintf(buf, "%s\r\n", command);
    write(fd, buf, strlen(command)+2);
    log_at_send(buf);

    /* read self input command */
    memset(buf, 0, MCTD_AT_BUF_SIZE);
    do {
        read_size = read_line(fd, buf);
        if (read_size < 0) {
            log_warn("read failed");
            sleep(1);
            return MCTD_RESULT_ERR_PROCESS;
        }

        if (read_size == 0) {
            retry_remain--;
            continue;
        }
        log_at_recv(buf);
    } while ((read_size <= 0) && (retry_remain > 0));

    if (retry_remain <= 0) {
        if (wait_second)
            *wait_second = 0;
        log_info("timed out.");
        sleep(1);
        return MCTD_RESULT_ERR_TIMEOUT;
    }

    /* read response */
    while(1) {
        if (!has_received_command() &&
                                (mctd_wait_data(fd, 1) != MCTD_RESULT_OK)) {
            retry_remain--;
            if (retry_remain == 0) {
                if (wait_second)
                    *wait_second = 0;
                log_info("timed out.");
                return MCTD_RESULT_ERR_TIMEOUT;
            }
            continue;
        }

        memset(buf, 0, MCTD_AT_BUF_SIZE);
        read_size = read_line(fd, buf);
        if (read_size < 0)
            break;

        if (read_size == 0) {
            retry_remain--;
            if (retry_remain == 0) {
                if (wait_second)
                    *wait_second = 0;
                log_info("timed out.");
                sleep(1);
                return MCTD_RESULT_ERR_TIMEOUT;
            }
            continue;
        }

        log_at_recv(buf);
        if (strstr(buf, "OK")) {
            result = MCTD_RESULT_OK;
            break;
        } else if(strstr(buf, "ERROR")) {
            result = MCTD_RESULT_ERR_PROCESS;
            break;
        } else {
            if (key && response && !(*response)) {
                if(strstr(buf, key)) {
                    *response = malloc(strlen(buf)+1);
                    if (!*response) {
                        perror("malloc");
                        result = MCTD_RESULT_ERR_MEMORY;
                        break;
                    }
                    memset(*response, 0, strlen(buf)+1);
                    strncpy(*response, buf, strlen(buf));
                }
            }
        }
    }

    if (wait_second)
        *wait_second = retry_remain;

    return result;
}

int mctd_atcmd_send_oneshot(const char *command, const char *key,
                                                            char **response)
{
    int fd;
    int result;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    result = mctd_atcmd_send(fd, command, key, response, NULL);

    mctd_serial_close(fd);

    return result;
}

int mctd_atcmd_send_and_echo_with_fd(int fd, const char *command,
                    char **response, bool check_result, int wait_second)
{
    int read_size;
    int total_size;
    char buf[MCTD_AT_BUF_SIZE];
    char *echo_buf = NULL;

    *response = malloc(1024);
    if (!*response) {
        perror("malloc");
        return MCTD_RESULT_ERR_MEMORY;
    }

    memset(*response, 0, 1024);

    echo_buf = *response;

    /* send command */
    sprintf(buf, "%s\r\n", command);
    write(fd, buf, strlen(command)+2);
    log_at_send(buf);

    memset(buf, 0, MCTD_AT_BUF_SIZE);
    read_size = read_line(fd, buf);
    if (read_size <= 0) {
        free(*response);
        *response = NULL;
        return MCTD_RESULT_ERR_PROCESS;
    }
    log_at_recv(buf);

    strncpy(echo_buf, buf, read_size);
    total_size = read_size;

    /* read response */
    while (has_received_command() ||
            (mctd_wait_data(fd, wait_second) == MCTD_RESULT_OK)) {
        memset(buf, 0, MCTD_AT_BUF_SIZE);
        read_size = read_line(fd, buf);
        if (read_size <= 0)
            break;

        if ((total_size + read_size) > 1024)
            break;

        strncpy(&(echo_buf[total_size]), buf, read_size);
        total_size += read_size;

        log_at_recv(buf);
        if (check_result == true) {
            if (strstr(buf, "OK"))
                break;
            else if(strstr(buf, "ERROR"))
                break;
        }
    }

    return MCTD_RESULT_OK;
}

int mctd_atcmd_send_and_echo_oneshot(const char *command, char **response)
{
    int fd;
    int result;

    fd = mctd_serial_open();
    if (fd < 0)
        return mctd_serial_convert_error(fd);

    result = mctd_atcmd_send_and_echo_with_fd(fd, command, response, true, 30);

    mctd_serial_close(fd);

    return result;
}

int mctd_atcmd_send_and_wait_with_fd(int fd, const char *command,
                                        const char *wait_str, int timeout_sec)
{
    int i;
    char buf[MCTD_AT_BUF_SIZE];
    int result;
    int read_size;
    int remain_second;

    remain_second = timeout_sec;

    result = mctd_atcmd_send(fd, command, NULL, NULL, &remain_second);
    if (result != MCTD_RESULT_OK)
        return result;

    result = MCTD_RESULT_ERR_TIMEOUT;
    for (i = 0; i < remain_second; i++) {
        if (!has_received_command() &&
                                    (mctd_wait_data(fd, 1) != MCTD_RESULT_OK))
            continue;

        /* wait response */
        memset(buf, 0, MCTD_AT_BUF_SIZE);
        read_size = read_line(fd, buf);
        if (read_size < 0)
            break;

        if (read_size == 0)
            continue;

        if (strstr(buf, wait_str)) {
            result = MCTD_RESULT_OK;
            break;
        }
    }

    return result;
}

int mctd_atcmd_send_and_wait_oneshot(const char *command,
                                        const char *wait_str, int timeout_sec)
{
    int fd;
    int result;

    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("open serial failed.");
        return mctd_serial_convert_error(fd);
    }

    result = mctd_atcmd_send_and_wait_with_fd(fd, command, wait_str,
                                                                timeout_sec);

    mctd_serial_close(fd);

    return result;
}

int mctd_atcmd_wait_urc_no_time_with_fd(int fd, const char *urc)
{
    int result;
    int read_size;
    char buf[MCTD_AT_BUF_SIZE];

    if (!urc)
        return MCTD_RESULT_ERR_PROCESS;

    result = MCTD_RESULT_ERR_TIMEOUT;
    if (has_received_command() ||
                    (mctd_wait_data_with_usec(fd, 0, 1) == MCTD_RESULT_OK)) {
        memset(buf, 0, MCTD_AT_BUF_SIZE);
        read_size = read_line(fd, buf);
        if (read_size > 0) {
            log_at_recv(buf);
            if (strstr(buf, urc))
                result = MCTD_RESULT_OK;
        }
    }

    return result;
}

#if 0 /* nobody used. */
int mctd_atcmd_wait_urc_no_time(const char *urc)
{
    int fd;
    int result;

    if (!urc)
        return MCTD_RESULT_ERR_PROCESS;

    fd = open_serial();
    if (fd < 0) {
        log_warn("open serial failed.");
        return mctd_serial_convert_error(fd);
    }

    result = mctd_atcmd_wait_urc_no_time_with_fd(fd, urc);
    close_serial(fd);
    return result;
}
#endif

int mctd_atcmd_wait_urc_with_fd_and_timeout(int fd, const char *urc,
                                                            int timeout_sec)
{
    int result;
    int i;
    int read_size;
    char buf[MCTD_AT_BUF_SIZE];

    if (!urc)
        return MCTD_RESULT_ERR_PROCESS;

    if (timeout_sec <= 0)
        return mctd_atcmd_wait_urc_no_time_with_fd(fd, urc);

    result = MCTD_RESULT_ERR_TIMEOUT;
    for (i = 0; i < timeout_sec; i++) {
        if ( !has_received_command() &&
                                    (mctd_wait_data(fd, 1) != MCTD_RESULT_OK))
            continue;

        memset(buf, 0, MCTD_AT_BUF_SIZE);
        read_size = read_line(fd, buf);
        if (read_size < 0) {
            result = MCTD_RESULT_ERR_SERIAL;
            break;
        }

        if (read_size == 0)
            continue;

        log_at_recv(buf);
        if (strstr(buf, urc)) {
            result = MCTD_RESULT_OK;
            break;
        }
    }

    return result;
}

#if 0 /* nobody used */
int mctd_atcmd_wait_urc_with_timeout(const char *urc, int timeout_sec)
{
    int fd;
    int result;

    if (!urc)
        return MCTD_RESULT_ERR_PROCESS;

    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("open serial failed.");
        return mctd_serial_convert_error(fd);
    }

    if (timeout_sec <= 0)
        result = mctd_atcmd_wait_urc_no_time_with_fd(fd, urc);
    else
        result = mctd_atcmd_wait_urc_with_fd_and_timeout(fd, urc, timeout_sec);

    mctd_serial_close(fd);
    return result;
}
#endif

void mctd_atcmd_flush()
{
    int fd;
    int read_size;
    char buf[MCTD_AT_BUF_SIZE];

    fd = mctd_serial_open();
    if (fd < 0) {
        log_warn("open serial failed.");
        return;
    }

    do {
        read_size = read_line_with_timeout(fd, buf, 0, 1);
        if (read_size > 0)
            log_at_recv(buf);
    } while(read_size > 0);

    mctd_serial_close(fd);
}
